@@ -8,13 +8,18 @@ import android.graphics.*
 import android.graphics.Bitmap
 import androidx.test.filters.MediumTest
 import androidx.test.ext.junit.runners.AndroidJUnit4
+import android.os.SystemClock
 import android.util.Base64
+import android.view.MotionEvent
 import java.io.ByteArrayOutputStream
 import org.hamcrest.Matchers.*
-import org.junit.Assert.fail
 import org.junit.Test
 import org.junit.runner.RunWith
 import org.mozilla.geckoview.GeckoResult
+import org.mozilla.geckoview.GeckoSession
+import org.mozilla.geckoview.GeckoSession.ContentDelegate
+import org.mozilla.geckoview.GeckoSession.ScrollDelegate
+import org.mozilla.geckoview.test.rule.GeckoSessionTestRule.AssertCalled
 import org.mozilla.geckoview.test.rule.GeckoSessionTestRule.WithDisplay
 import org.hamcrest.Matchers.closeTo
 import org.hamcrest.Matchers.equalTo
@@ -149,15 +154,15 @@ class DynamicToolbarTest : BaseSessionTest() {
         mainSession.loadTestPath(BaseSessionTest.FIXED_VH)
         mainSession.waitForPageStop()
 
-        val pixelRatio = sessionRule.session.evaluateJS("window.devicePixelRatio") as Double
-        val scale = sessionRule.session.evaluateJS("window.visualViewport.scale") as Double
+        val pixelRatio = mainSession.evaluateJS("window.devicePixelRatio") as Double
+        val scale = mainSession.evaluateJS("window.visualViewport.scale") as Double
 
         for (i in 1..dynamicToolbarMaxHeight) {
           // Simulate the dynamic toolbar is going to be hidden.
           sessionRule.display?.run { setVerticalClipping(-i) }
 
           val expectedViewportHeight = (SCREEN_HEIGHT - dynamicToolbarMaxHeight + i) / scale / pixelRatio
-          val promise = sessionRule.session.evaluatePromiseJS("""
+          val promise = mainSession.evaluatePromiseJS("""
              new Promise(resolve => {
                window.visualViewport.addEventListener('resize', resolve(window.visualViewport.height));
              });
@@ -200,7 +205,7 @@ class DynamicToolbarTest : BaseSessionTest() {
             getComputedStyle(document.querySelector('#fixed-element')).height
         """.trimIndent()) as String
 
-        val scale = sessionRule.session.evaluateJS("window.visualViewport.scale") as Double
+        val scale = mainSession.evaluateJS("window.visualViewport.scale") as Double
         val expectedHeight = (SCREEN_HEIGHT / scale).toInt()
         assertThat("The %-based height should be now recomputed based on the screen height",
                    height, equalTo(expectedHeight.toString() + "px"))
@@ -219,7 +224,7 @@ class DynamicToolbarTest : BaseSessionTest() {
         mainSession.waitForPageStop()
 
         for (i in 1..dynamicToolbarMaxHeight - 1) {
-            val promise = sessionRule.session.evaluatePromiseJS("""
+            val promise = mainSession.evaluatePromiseJS("""
                 new Promise(resolve => {
                     let fired = false;
                     window.addEventListener('resize', () => { fired = true; }, { once: true });
@@ -236,7 +241,7 @@ class DynamicToolbarTest : BaseSessionTest() {
                        promise.value as Boolean, equalTo(false));
         }
 
-        val promise = sessionRule.session.evaluatePromiseJS("""
+        val promise = mainSession.evaluatePromiseJS("""
             new Promise(resolve => {
                 window.addEventListener('resize', () => { resolve(true); }, { once: true });
             });
@@ -262,10 +267,10 @@ class DynamicToolbarTest : BaseSessionTest() {
         mainSession.loadTestPath(BaseSessionTest.FIXED_BOTTOM)
         mainSession.waitForPageStop()
 
-        val pixelRatio = sessionRule.session.evaluateJS("window.devicePixelRatio") as Double
+        val pixelRatio = mainSession.evaluateJS("window.devicePixelRatio") as Double
 
         for (i in 1..dynamicToolbarMaxHeight - 1) {
-            val promise = sessionRule.session.evaluatePromiseJS("""
+            val promise = mainSession.evaluatePromiseJS("""
                new Promise(resolve => {
                  window.visualViewport.addEventListener('resize', resolve(window.innerHeight));
                });
@@ -277,7 +282,7 @@ class DynamicToolbarTest : BaseSessionTest() {
                        promise.value as Double, closeTo(SCREEN_HEIGHT / 2 / pixelRatio, .01))
         }
 
-        val promise = sessionRule.session.evaluatePromiseJS("""
+        val promise = mainSession.evaluatePromiseJS("""
             new Promise(resolve => {
                 window.addEventListener('resize', () => { resolve(window.innerHeight); }, { once: true });
             });
@@ -300,7 +305,7 @@ class DynamicToolbarTest : BaseSessionTest() {
         mainSession.loadTestPath(BaseSessionTest.FIXED_VH)
         mainSession.waitForPageStop()
 
-        val promise = sessionRule.session.evaluatePromiseJS("""
+        val promise = mainSession.evaluatePromiseJS("""
             new Promise(resolve => window.addEventListener('resize', () => resolve(true)));
         """.trimIndent())
 
@@ -311,4 +316,167 @@ class DynamicToolbarTest : BaseSessionTest() {
 
         assertThat("Got a rezie event", promise.value as Boolean, equalTo(true))
     }
+
+    @WithDisplay(height = SCREEN_HEIGHT, width = SCREEN_WIDTH)
+    @Test fun showDynamicToolbar() {
+        val dynamicToolbarMaxHeight = SCREEN_HEIGHT / 2
+        sessionRule.display?.run { setDynamicToolbarMaxHeight(dynamicToolbarMaxHeight) }
+
+        // Set active since setVerticalClipping call affects only for forground tab.
+        mainSession.setActive(true)
+
+        mainSession.loadTestPath(SHOW_DYNAMIC_TOOLBAR_HTML_PATH)
+        mainSession.waitForPageStop()
+        mainSession.evaluateJS("window.scrollTo(0, " + dynamicToolbarMaxHeight + ")")
+        mainSession.waitUntilCalled(object : ScrollDelegate {
+            @AssertCalled(count = 1)
+            override fun onScrollChanged(session: GeckoSession, scrollX: Int, scrollY: Int) {
+            }
+        })
+
+        // Simulate the dynamic toolbar being hidden by the scroll
+        sessionRule.display?.run { setVerticalClipping(-dynamicToolbarMaxHeight) }
+
+        mainSession.synthesizeTap(5, 25)
+
+        mainSession.waitUntilCalled(object : ContentDelegate {
+            @AssertCalled(count = 1)
+            override fun onShowDynamicToolbar(session: GeckoSession) {
+            }
+        })
+    }
+
+    @WithDisplay(height = SCREEN_HEIGHT, width = SCREEN_WIDTH)
+    @Test fun showDynamicToolbarOnOverflowHidden() {
+        val dynamicToolbarMaxHeight = SCREEN_HEIGHT / 2
+        sessionRule.display?.run { setDynamicToolbarMaxHeight(dynamicToolbarMaxHeight) }
+
+        // Set active since setVerticalClipping call affects only for forground tab.
+        mainSession.setActive(true)
+
+        mainSession.loadTestPath(SHOW_DYNAMIC_TOOLBAR_HTML_PATH)
+        mainSession.waitForPageStop()
+        mainSession.evaluateJS("window.scrollTo(0, " + dynamicToolbarMaxHeight + ")")
+        mainSession.waitUntilCalled(object : ScrollDelegate {
+            @AssertCalled(count = 1)
+            override fun onScrollChanged(session: GeckoSession, scrollX: Int, scrollY: Int) {
+            }
+        })
+
+        // Simulate the dynamic toolbar being hidden by the scroll
+        sessionRule.display?.run { setVerticalClipping(-dynamicToolbarMaxHeight) }
+
+        mainSession.evaluateJS("document.documentElement.style.overflow = 'hidden'")
+
+        mainSession.waitUntilCalled(object : ContentDelegate {
+            @AssertCalled(count = 1)
+            override fun onShowDynamicToolbar(session: GeckoSession) {
+            }
+        })
+    }
+
+    private fun getComputedViewportHeight(style: String): Double {
+        val viewportHeight = mainSession.evaluateJS("""
+            const target = document.createElement('div');
+            target.style.height = '$style';
+            document.body.appendChild(target);
+            parseFloat(getComputedStyle(target).height);
+        """.trimIndent()) as Double
+
+        return viewportHeight
+    }
+
+    @WithDisplay(height = SCREEN_HEIGHT, width = SCREEN_WIDTH)
+    @Test
+    fun viewportVariants() {
+        val dynamicToolbarMaxHeight = SCREEN_HEIGHT / 2
+        sessionRule.display?.run { setDynamicToolbarMaxHeight(dynamicToolbarMaxHeight) }
+
+        // Set active since setVerticalClipping call affects only for forground tab.
+        mainSession.setActive(true)
+
+        mainSession.loadTestPath(BaseSessionTest.VIEWPORT_PATH)
+        mainSession.waitForPageStop()
+
+        val pixelRatio = mainSession.evaluateJS("window.devicePixelRatio") as Double
+        val scale = mainSession.evaluateJS("window.visualViewport.scale") as Double
+
+        var smallViewportHeight = getComputedViewportHeight("100svh");
+        assertThat("svh value at the initial state", smallViewportHeight,
+                   closeTo((SCREEN_HEIGHT - dynamicToolbarMaxHeight) / scale / pixelRatio, 0.1));
+
+        var largeViewportHeight = getComputedViewportHeight("100lvh");
+        assertThat("lvh value at the initial state", largeViewportHeight,
+                   closeTo(SCREEN_HEIGHT / scale / pixelRatio, 0.1));
+
+        var dynamicViewportHeight = getComputedViewportHeight("100dvh");
+        assertThat("dvh value at the initial state", dynamicViewportHeight,
+                   closeTo((SCREEN_HEIGHT - dynamicToolbarMaxHeight) / scale / pixelRatio, 0.1));
+
+        // Move down the toolbar at a fourth of its position.
+        sessionRule.display?.run { setVerticalClipping(-dynamicToolbarMaxHeight / 4) }
+
+        smallViewportHeight = getComputedViewportHeight("100svh");
+        assertThat("svh value during toolbar transition", smallViewportHeight,
+                   closeTo((SCREEN_HEIGHT - dynamicToolbarMaxHeight) / scale / pixelRatio, 0.1));
+
+        largeViewportHeight = getComputedViewportHeight("100lvh");
+        assertThat("lvh value during toolbar transition", largeViewportHeight,
+                   closeTo(SCREEN_HEIGHT / scale / pixelRatio, 0.1));
+
+        dynamicViewportHeight = getComputedViewportHeight("100dvh");
+        assertThat("dvh value during toolbar transition", dynamicViewportHeight,
+                   closeTo((SCREEN_HEIGHT - dynamicToolbarMaxHeight + dynamicToolbarMaxHeight / 4) / scale / pixelRatio, 0.1));
+    }
+
+    // With dynamic toolbar, there was a floating point rounding error in Gecko layout side.
+    // The error was appeared by user interactive async scrolling, not by programatic async
+    // scrolling, e.g. scrollTo() method. If the error happens there will appear 1px gap
+    // between <body> and an element which covers up the <body> element.
+    // This test simulates the situation.
+    @WithDisplay(height = SCREEN_HEIGHT, width = SCREEN_WIDTH)
+    @Test
+    fun noGapAppearsBetweenBodyAndElementFullyCoveringBody() {
+        val dynamicToolbarMaxHeight = SCREEN_HEIGHT / 2
+        sessionRule.display?.run { setDynamicToolbarMaxHeight(dynamicToolbarMaxHeight) }
+
+        // Set active since setVerticalClipping call affects only for forground tab.
+        mainSession.setActive(true)
+
+        val reference = getComparisonScreenshot(SCREEN_WIDTH, SCREEN_HEIGHT)
+
+        mainSession.loadTestPath(BaseSessionTest.BODY_FULLY_COVERED_BY_GREEN_ELEMENT)
+        mainSession.waitForPageStop()
+        mainSession.flushApzRepaints()
+
+        // Scrolling down by touch events.
+        var downTime = SystemClock.uptimeMillis();
+        var down = MotionEvent.obtain(
+                downTime, SystemClock.uptimeMillis(), MotionEvent.ACTION_DOWN, 50f, 70f, 0)
+        mainSession.panZoomController.onTouchEvent(down)
+        var move = MotionEvent.obtain(
+                downTime, SystemClock.uptimeMillis(), MotionEvent.ACTION_MOVE, 50f, 30f, 0)
+        mainSession.panZoomController.onTouchEvent(move)
+        var up = MotionEvent.obtain(
+                downTime, SystemClock.uptimeMillis(), MotionEvent.ACTION_UP, 50f, 10f, 0)
+        mainSession.panZoomController.onTouchEvent(up)
+        mainSession.flushApzRepaints()
+
+        // Scrolling up by touch events to restore the original position.
+        downTime = SystemClock.uptimeMillis();
+        down = MotionEvent.obtain(
+                downTime, SystemClock.uptimeMillis(), MotionEvent.ACTION_DOWN, 50f, 10f, 0)
+        mainSession.panZoomController.onTouchEvent(down)
+        move = MotionEvent.obtain(
+                downTime, SystemClock.uptimeMillis(), MotionEvent.ACTION_MOVE, 50f, 30f, 0)
+        mainSession.panZoomController.onTouchEvent(move)
+        up = MotionEvent.obtain(
+                downTime, SystemClock.uptimeMillis(), MotionEvent.ACTION_UP, 50f, 70f, 0)
+        mainSession.panZoomController.onTouchEvent(up)
+        mainSession.flushApzRepaints()
+
+        sessionRule.display?.let {
+            assertScreenshotResult(it.capturePixels(), reference)
+        }
+    }
 }
