@@ -4,6 +4,7 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
+#include "ErrorList.h"
 #include "nsError.h"
 #include "mozilla/CheckedInt.h"
 #include "mozilla/Likely.h"
@@ -31,7 +32,7 @@ nsHtml5TreeBuilder::nsHtml5TreeBuilder(nsHtml5OplessBuilder* aBuilder)
       headPointer(nullptr),
       charBufferLen(0),
       quirks(false),
-      isSrcdocDocument(false),
+      forceNoQuirks(false),
       mBuilder(aBuilder),
       mViewSource(nullptr),
       mOpSink(nullptr),
@@ -40,7 +41,8 @@ nsHtml5TreeBuilder::nsHtml5TreeBuilder(nsHtml5OplessBuilder* aBuilder)
       mSpeculativeLoadStage(nullptr),
       mBroken(NS_OK),
       mCurrentHtmlScriptIsAsyncOrDefer(false),
-      mPreventScriptExecution(false)
+      mPreventScriptExecution(false),
+      mGenerateSpeculativeLoads(false)
 #ifdef DEBUG
       ,
       mActive(false)
@@ -50,7 +52,8 @@ nsHtml5TreeBuilder::nsHtml5TreeBuilder(nsHtml5OplessBuilder* aBuilder)
 }
 
 nsHtml5TreeBuilder::nsHtml5TreeBuilder(nsAHtml5TreeOpSink* aOpSink,
-                                       nsHtml5TreeOpStage* aStage)
+                                       nsHtml5TreeOpStage* aStage,
+                                       bool aGenerateSpeculativeLoads)
     : mode(0),
       originalMode(0),
       framesetOk(false),
@@ -70,7 +73,7 @@ nsHtml5TreeBuilder::nsHtml5TreeBuilder(nsAHtml5TreeOpSink* aOpSink,
       headPointer(nullptr),
       charBufferLen(0),
       quirks(false),
-      isSrcdocDocument(false),
+      forceNoQuirks(false),
       mBuilder(nullptr),
       mViewSource(nullptr),
       mOpSink(aOpSink),
@@ -79,12 +82,15 @@ nsHtml5TreeBuilder::nsHtml5TreeBuilder(nsAHtml5TreeOpSink* aOpSink,
       mSpeculativeLoadStage(aStage),
       mBroken(NS_OK),
       mCurrentHtmlScriptIsAsyncOrDefer(false),
-      mPreventScriptExecution(false)
+      mPreventScriptExecution(false),
+      mGenerateSpeculativeLoads(aGenerateSpeculativeLoads)
 #ifdef DEBUG
       ,
       mActive(false)
 #endif
 {
+  MOZ_ASSERT(!(!aStage && aGenerateSpeculativeLoads),
+             "Must not generate speculative loads without a stage");
   MOZ_COUNT_CTOR(nsHtml5TreeBuilder);
 }
 
@@ -167,7 +173,7 @@ nsIContentHandle* nsHtml5TreeBuilder::createElement(
 
   // Start wall of code for speculative loading and line numbers
 
-  if (mSpeculativeLoadStage && mode != IN_TEMPLATE) {
+  if (mGenerateSpeculativeLoads && mode != IN_TEMPLATE) {
     switch (aNamespace) {
       case kNameSpaceID_XHTML:
         if (nsGkAtoms::img == aName) {
@@ -986,12 +992,15 @@ void nsHtml5TreeBuilder::elementPushed(int32_t aNamespace, nsAtom* aName,
     }
     return;
   }
-  if (mSpeculativeLoadStage && aName == nsGkAtoms::picture) {
-    // mSpeculativeLoadStage is non-null only in the off-the-main-thread
-    // tree builder, which handles the network stream
-    //
+  if (mGenerateSpeculativeLoads && aName == nsGkAtoms::picture) {
     // See comments in nsHtml5SpeculativeLoad.h about <picture> preloading
     mSpeculativeLoadQueue.AppendElement()->InitOpenPicture();
+    return;
+  }
+  if (aName == nsGkAtoms::_template) {
+    if (tokenizer->TemplatePushedOrHeadPopped()) {
+      requestSuspension();
+    }
   }
 }
 
@@ -1059,6 +1068,11 @@ void nsHtml5TreeBuilder::elementPopped(int32_t aNamespace, nsAtom* aName,
     }
     opDoneAddingChildren operation(aElement);
     treeOp->Init(mozilla::AsVariant(operation));
+    if (aNamespace == kNameSpaceID_XHTML && aName == nsGkAtoms::head) {
+      if (tokenizer->TemplatePushedOrHeadPopped()) {
+        requestSuspension();
+      }
+    }
     return;
   }
   if (aName == nsGkAtoms::style ||
@@ -1106,12 +1120,10 @@ void nsHtml5TreeBuilder::elementPopped(int32_t aNamespace, nsAtom* aName,
     return;
   }
 
-  if (mSpeculativeLoadStage && aName == nsGkAtoms::picture) {
-    // mSpeculativeLoadStage is non-null only in the off-the-main-thread
-    // tree builder, which handles the network stream
-    //
+  if (mGenerateSpeculativeLoads && aName == nsGkAtoms::picture) {
     // See comments in nsHtml5SpeculativeLoad.h about <picture> preloading
     mSpeculativeLoadQueue.AppendElement()->InitEndPicture();
+    return;
   }
 }
 
@@ -1190,7 +1202,7 @@ bool nsHtml5TreeBuilder::HasScript() {
   return mOpQueue.ElementAt(len - 1).IsRunScript();
 }
 
-bool nsHtml5TreeBuilder::Flush(bool aDiscretionary) {
+mozilla::Result<bool, nsresult> nsHtml5TreeBuilder::Flush(bool aDiscretionary) {
   if (MOZ_UNLIKELY(mBuilder)) {
     MOZ_ASSERT_UNREACHABLE("Must never flush with builder.");
     return false;
@@ -1219,7 +1231,9 @@ bool nsHtml5TreeBuilder::Flush(bool aDiscretionary) {
                    "Tree builder is broken but the op in queue is not marked "
                    "as broken.");
       }
-      mOpSink->MoveOpsFrom(mOpQueue);
+      if (!mOpSink->MoveOpsFrom(mOpQueue)) {
+        return Err(NS_ERROR_OUT_OF_MEMORY);
+      }
     }
     return hasOps;
   }
@@ -1239,16 +1253,26 @@ void nsHtml5TreeBuilder::FlushLoads() {
 }
 
 void nsHtml5TreeBuilder::SetDocumentCharset(NotNull<const Encoding*> aEncoding,
-                                            int32_t aCharsetSource) {
-  if (mBuilder) {
-    mBuilder->SetDocumentCharsetAndSource(aEncoding, aCharsetSource);
-  } else if (mSpeculativeLoadStage) {
-    mSpeculativeLoadQueue.AppendElement()->InitSetDocumentCharset(
-        aEncoding, aCharsetSource);
-  } else {
-    opSetDocumentCharset opearation(aEncoding, aCharsetSource);
-    mOpQueue.AppendElement()->Init(mozilla::AsVariant(opearation));
+                                            nsCharsetSource aCharsetSource,
+                                            bool aCommitEncodingSpeculation) {
+  MOZ_ASSERT(!mBuilder, "How did we call this with builder?");
+  MOZ_ASSERT(mSpeculativeLoadStage,
+             "How did we call this without a speculative load stage?");
+  mSpeculativeLoadQueue.AppendElement()->InitSetDocumentCharset(
+      aEncoding, aCharsetSource, aCommitEncodingSpeculation);
+}
+
+void nsHtml5TreeBuilder::UpdateCharsetSource(nsCharsetSource aCharsetSource) {
+  MOZ_ASSERT(!mBuilder, "How did we call this with builder?");
+  MOZ_ASSERT(mSpeculativeLoadStage,
+             "How did we call this without a speculative load stage (even "
+             "though we don't need it right here)?");
+  if (mViewSource) {
+    mViewSource->UpdateCharsetSource(aCharsetSource);
+    return;
   }
+  opUpdateCharsetSource operation(aCharsetSource);
+  mOpQueue.AppendElement()->Init(mozilla::AsVariant(operation));
 }
 
 void nsHtml5TreeBuilder::StreamEnded() {
@@ -1285,9 +1309,15 @@ void nsHtml5TreeBuilder::MaybeComplainAboutCharset(const char* aMsgId,
     MOZ_ASSERT_UNREACHABLE("Must never complain about charset with builder.");
     return;
   }
-  opMaybeComplainAboutCharset opeartion(const_cast<char*>(aMsgId), aError,
-                                        aLineNumber);
-  mOpQueue.AppendElement()->Init(mozilla::AsVariant(opeartion));
+
+  if (mSpeculativeLoadStage) {
+    mSpeculativeLoadQueue.AppendElement()->InitMaybeComplainAboutCharset(
+        aMsgId, aError, aLineNumber);
+  } else {
+    opMaybeComplainAboutCharset opeartion(const_cast<char*>(aMsgId), aError,
+                                          aLineNumber);
+    mOpQueue.AppendElement()->Init(mozilla::AsVariant(opeartion));
+  }
 }
 
 void nsHtml5TreeBuilder::TryToEnableEncodingMenu() {
@@ -1365,6 +1395,7 @@ void nsHtml5TreeBuilder::StartPlainTextViewSource(const nsAutoString& aTitle) {
 
 void nsHtml5TreeBuilder::StartPlainText() {
   MOZ_ASSERT(!mBuilder, "Must not view source with builder.");
+  setForceNoQuirks(true);
   startTag(nsHtml5ElementName::ELT_LINK,
            nsHtml5PlainTextUtils::NewLinkAttributes(), false);
 
@@ -1504,13 +1535,13 @@ void nsHtml5TreeBuilder::errStrayDoctype() {
 }
 
 void nsHtml5TreeBuilder::errAlmostStandardsDoctype() {
-  if (MOZ_UNLIKELY(mViewSource) && !isSrcdocDocument) {
+  if (MOZ_UNLIKELY(mViewSource) && !forceNoQuirks) {
     mViewSource->AddErrorToCurrentRun("errAlmostStandardsDoctype");
   }
 }
 
 void nsHtml5TreeBuilder::errQuirkyDoctype() {
-  if (MOZ_UNLIKELY(mViewSource) && !isSrcdocDocument) {
+  if (MOZ_UNLIKELY(mViewSource) && !forceNoQuirks) {
     mViewSource->AddErrorToCurrentRun("errQuirkyDoctype");
   }
 }
@@ -1558,7 +1589,7 @@ void nsHtml5TreeBuilder::errFooBetweenHeadAndBody(nsAtom* aName) {
 }
 
 void nsHtml5TreeBuilder::errStartTagWithoutDoctype() {
-  if (MOZ_UNLIKELY(mViewSource) && !isSrcdocDocument) {
+  if (MOZ_UNLIKELY(mViewSource) && !forceNoQuirks) {
     mViewSource->AddErrorToCurrentRun("errStartTagWithoutDoctype");
   }
 }
@@ -1648,7 +1679,7 @@ void nsHtml5TreeBuilder::errStartTagInTableBody(nsAtom* aName) {
 }
 
 void nsHtml5TreeBuilder::errEndTagSeenWithoutDoctype() {
-  if (MOZ_UNLIKELY(mViewSource) && !isSrcdocDocument) {
+  if (MOZ_UNLIKELY(mViewSource) && !forceNoQuirks) {
     mViewSource->AddErrorToCurrentRun("errEndTagSeenWithoutDoctype");
   }
 }
@@ -1745,3 +1776,9 @@ void nsHtml5TreeBuilder::errEndWithUnclosedElements(nsAtom* aName) {
     mViewSource->AddErrorToCurrentRun("errEndWithUnclosedElements", aName);
   }
 }
+
+void nsHtml5TreeBuilder::errListUnclosedStartTags(int32_t aIgnored) {
+  if (MOZ_UNLIKELY(mViewSource)) {
+    mViewSource->AddErrorToCurrentRun("errListUnclosedStartTags");
+  }
+}
