@@ -22,6 +22,7 @@
 #include "nsHashKeys.h"
 #include "mozilla/LinkedList.h"
 #include "nsHtml5DocumentBuilder.h"
+#include "nsCharsetSource.h"
 
 class nsHtml5Parser;
 class nsHtml5StreamParser;
@@ -110,10 +111,10 @@ class nsHtml5TreeOpExecutor final
    */
   NS_IMETHOD WillParse() override;
 
-  /**
-   *
-   */
-  NS_IMETHOD WillBuildModel(nsDTDMode aDTDMode) override;
+  NS_IMETHOD WillBuildModel(nsDTDMode /* unused */) override {
+    return WillBuildModel();
+  }
+  nsresult WillBuildModel();
 
   /**
    * Emits EOF.
@@ -128,7 +129,7 @@ class nsHtml5TreeOpExecutor final
   /**
    * Unimplemented. For interface compat only.
    */
-  NS_IMETHOD WillResume() override;
+  void WillResume() override;
 
   virtual void InitialTranslationCompleted() override;
 
@@ -181,17 +182,27 @@ class nsHtml5TreeOpExecutor final
 
   nsresult FlushDocumentWrite();
 
+  void CommitToInternalEncoding();
+
+  [[nodiscard]] bool TakeOpsFromStage();
+
   void MaybeSuspend();
 
   void Start();
 
+  void SetDocumentCharsetAndSource(NotNull<const Encoding*> aEncoding,
+                                   nsCharsetSource aCharsetSource);
+
+  void UpdateCharsetSource(nsCharsetSource aCharsetSource);
+
   void NeedsCharsetSwitchTo(NotNull<const Encoding*> aEncoding, int32_t aSource,
                             uint32_t aLineNumber);
 
   void MaybeComplainAboutCharset(const char* aMsgId, bool aError,
                                  uint32_t aLineNumber);
 
-  void ComplainAboutBogusProtocolCharset(mozilla::dom::Document*);
+  void ComplainAboutBogusProtocolCharset(mozilla::dom::Document* aDoc,
+                                         bool aUnrecognized);
 
   void MaybeComplainAboutDeepTree(uint32_t aLineNumber);
 
@@ -209,7 +220,8 @@ class nsHtml5TreeOpExecutor final
    * Flush the operations from the tree operations from the argument
    * queue unconditionally. (This is for the main thread case.)
    */
-  virtual void MoveOpsFrom(nsTArray<nsHtml5TreeOperation>& aOpQueue) override;
+  [[nodiscard]] virtual bool MoveOpsFrom(
+      nsTArray<nsHtml5TreeOperation>& aOpQueue) override;
 
   void ClearOpQueue();
 
